#include "menudisplay.h"

// local includes
#include "tftinstance.h"

#include "esp_log.h"

using namespace std::chrono_literals;

namespace espgui {
void MenuDisplay::start()
{
    Base::start();

//    m_selectedIndex = 0;
//    m_scrollOffset = 0;

    m_rotateOffset = 0;
    m_pressed = false;

    m_upScrolling = false;
    m_upHoldingSince = std::nullopt;

    m_downScrolling = false;
    m_downHoldingSince = std::nullopt;

    m_leftScrolling = false;
    m_leftHoldingSince = std::nullopt;

    m_rightScrolling = false;
    m_rightHoldingSince = std::nullopt;
}

void MenuDisplay::initScreen()
{
    Base::initScreen();

    for (auto &label : m_labels)
        label.start();

    runForEveryMenuItem([](MenuItem &item){
        item.start();
    });

    m_icons.fill(nullptr);

    m_highlightedIndex = -1;
}

void MenuDisplay::update()
{
    Base::update();

    {
        const auto ago = espchrono::ago(*m_upHoldingSince);
        if (m_upHoldingSince && (m_upScrolling || ago > getInitialScrollSpeed()))
        {
            if (ago > getScrollSpeed())
            {
                if (!m_upScrolling)
                {
                    // gets executed one time
                    buttonHeld(espgui::Up);
                }
                m_upScrolling = true;
                m_upHoldingSince = espchrono::millis_clock::now();

                // do sth
                m_rotateOffset--;
            }
        }
    }

    {
        const auto ago = espchrono::ago(*m_downHoldingSince);
        if (m_downHoldingSince && (m_downScrolling || ago > getInitialScrollSpeed()))
        {
            if (ago > getScrollSpeed())
            {
                if (!m_downScrolling)
                {
                    // gets executed one time
                    buttonHeld(espgui::Down);
                }
                m_downScrolling = true;
                m_downHoldingSince = espchrono::millis_clock::now();

                // do sth
                m_rotateOffset++;
            }
        }
    }

    {
        const auto ago = espchrono::ago(*m_leftHoldingSince);
        if (m_leftHoldingSince && (m_leftScrolling || ago > getInitialScrollSpeed()))
        {
            if (ago > getScrollSpeed())
            {
                if (!m_leftScrolling)
                {
                    // gets executed one time
                    buttonHeld(espgui::Left);
                }
                m_leftScrolling = true;
                m_leftHoldingSince = espchrono::millis_clock::now();

                // do sth
            }
        }
    }

    {
        const auto ago = espchrono::ago(*m_rightHoldingSince);
        if (m_rightHoldingSince && (m_rightScrolling || ago > getInitialScrollSpeed()))
        {
            if (ago > getScrollSpeed())
            {
                if (!m_rightScrolling)
                {
                    // gets executed one time
                    buttonHeld(espgui::Right);
                }
                m_rightScrolling = true;
                m_rightHoldingSince = espchrono::millis_clock::now();

                // do sth
            }
        }
    }

    if (!m_pressed)
    {
        const auto offset = m_rotateOffset;
        m_rotateOffset = 0;

        const auto itemCount = menuItemCount();

        if (itemCount)
        {
            if (m_selectedIndex == -1)
                m_selectedIndex = 0;

            m_selectedIndex = m_selectedIndex + offset;

            if (m_selectedIndex < 0)
                m_selectedIndex += itemCount;
            if (m_selectedIndex >= itemCount)
                m_selectedIndex -= itemCount;

            if (m_selectedIndex < m_scrollOffset)
                m_scrollOffset = m_selectedIndex;
            if (m_selectedIndex >= m_scrollOffset + m_labels.size())
                m_scrollOffset = m_selectedIndex - m_labels.size() + 1;
        }
        else
        {
            m_selectedIndex = -1;
            m_scrollOffset = 0;
        }

        runForEveryMenuItem([&](MenuItem &item){
            item.update();
        });

        if (m_selectedIndex >= 0 && getMenuItem(m_selectedIndex).skipScroll())
        {
            if (offset > 0)
            {
                m_rotateOffset++;
            }
            else if (offset < 0)
            {
                m_rotateOffset--;
            }
        }
    }
    else
    {
        m_pressed = false;
        if (m_selectedIndex >= 0)
            getMenuItem(m_selectedIndex).triggered();
    }
}

void MenuDisplay::redraw()
{
    Base::redraw();

    tft.setTextFont(4);
    tft.setTextColor(TFT_YELLOW, TFT_BLACK);

    int i{0};

    auto labelsIter = std::begin(m_labels);

    auto iconsIter = std::begin(m_icons);

    int newHighlightedIndex{-1};

    const auto drawItemRect = [](const auto &label, const auto color){
        tft.drawRect(5,
                     label.y()-1,
                     tft.width() - 10,
                     lineHeight+1,
                     color);
    };

    runForEveryMenuItem([&](MenuItem &item){
        const auto index = i++;

        if (!item.visible())
        {
            return;
        }

        if (index < m_scrollOffset)
            return;

        if (labelsIter == std::end(m_labels))
            return;

        const auto relativeIndex = index - m_scrollOffset;
        const auto selected = index == m_selectedIndex;

        if (selected)
            newHighlightedIndex = relativeIndex;
        else if (relativeIndex == m_highlightedIndex)
            drawItemRect(*labelsIter, TFT_BLACK);

        tft.setTextFont(item.font());
        tft.setTextColor(item.color(), TFT_BLACK);
        labelsIter->redraw(item.text());

        if (item.icon() != *iconsIter)
        {
            tft.fillRect(6, labelsIter->y()+1, 24, 24, TFT_BLACK);

            auto icon = item.icon();
            if (icon)
            {
                tft.setSwapBytes(true);
                tft.pushImage(6, labelsIter->y()+1, icon->WIDTH, icon->HEIGHT, icon->buffer);
                tft.setSwapBytes(false);
            }
            *iconsIter = icon;
        }

        if (selected && (relativeIndex != m_highlightedIndex))
        {
            drawItemRect(*labelsIter, TFT_WHITE);
        }

        labelsIter++;
        iconsIter++;
    });

    for (; labelsIter != std::end(m_labels); labelsIter++, iconsIter++)
    {
        const auto relativeIndex = std::distance(std::begin(m_labels), labelsIter);

        if (relativeIndex == m_highlightedIndex)
            drawItemRect(*labelsIter, TFT_BLACK);

        labelsIter->clear();

        if (*iconsIter)
        {
            tft.fillRect(5, labelsIter->y()+1, 24, 24, TFT_BLACK);
            *iconsIter = nullptr;
        }
    }

    m_highlightedIndex = newHighlightedIndex;
}

void MenuDisplay::stop()
{
    Base::stop();

    runForEveryMenuItem([](MenuItem &item){
        item.stop();
    });
}

void MenuDisplay::buttonPressed(Button button)
{
    //Base::buttonPressed(button);

    switch (button)
    {
    case Button::Left: this->back(); m_leftHoldingSince = espchrono::millis_clock::now(); m_leftScrolling = false; break;
    case Button::Right: m_pressed = true; m_rightHoldingSince = espchrono::millis_clock::now(); m_rightScrolling = false; break;
    case Button::Up: m_rotateOffset--; m_upHoldingSince = espchrono::millis_clock::now(); m_upScrolling = false; break;
    case Button::Down: m_rotateOffset++; m_downHoldingSince = espchrono::millis_clock::now(); m_downScrolling = false; break;
    }
}

void MenuDisplay::buttonReleased(Button button)
{
    //Base::buttonPressed(button);

    switch (button)
    {
    case Button::Up: m_upHoldingSince = std::nullopt; break;
    case Button::Down: m_downHoldingSince = std::nullopt; break;
    case Button::Left: m_leftHoldingSince = std::nullopt; break;
    case Button::Right: m_rightHoldingSince = std::nullopt; break;
    default:;
    }
}

void MenuDisplay::buttonHeld(Button button)
{
    ESP_LOGI("Menudisplay", "buttonHeld %d", button);
}
} // namespace espgui
